from __future__ import annotations

from pathlib import Path
from typing import Callable, Union

from kraken.core import Project, Property, Task, TaskResult

DEFAULT_ENCODING = "utf-8"


class RenderFileTask(Task):
    content: Property[Union[str, Callable[[], str]]]
    file: Property[Path]
    encoding: Property[str]

    def __init__(self, name: str, project: Project) -> None:
        super().__init__(name, project)
        self.encoding.setdefault(DEFAULT_ENCODING)

    def _get_content(self) -> str:
        content = self.content.get()
        if callable(content):
            content = content()
        return content

    def finalize(self) -> None:
        # Materialize content that is to be rendered.
        content = self._get_content()
        self.content.setfinal(content)
        return super().finalize()

    def is_up_to_date(self) -> bool:
        content = self._get_content()
        return self.file.get().is_file() and self.file.get().read_text(self.encoding.get()) == content

    def execute(self) -> TaskResult:
        content = self._get_content()
        self.file.get().parent.mkdir(exist_ok=True)
        encoded = content.encode(self.encoding.get())
        print(f"write {self.file} ({len(encoded)} bytes)")
        self.file.get().write_bytes(encoded)
        return TaskResult.SUCCEEDED



from __future__ import annotations

import base64
import contextlib
import json
import shlex
import tempfile
from pathlib import Path

from kraken.core.project import Project
from kraken.core.property import Property
from kraken.core.task import TaskResult
from kraken.core.utils import flatten

from . import DockerBuildTask
from .dockerapi import docker_load, docker_run


class KanikoBuildTask(DockerBuildTask):
    """An implementation for building Docker images with Kaniko."""

    kaniko_image: Property[str]
    kaniko_context: Property[str]
    kaniko_cache_copy_layers: Property[bool]
    kaniko_snapshot_mode: Property[str]
    kaniko_secrets_mount_dir: Property[str]

    def __init__(self, name: str, project: Project) -> None:
        super().__init__(name, project)
        self.kaniko_image.set("gcr.io/kaniko-project/executor:debug")
        self.kaniko_context.set("/workspace")
        self.kaniko_cache_copy_layers.set(True)
        self.kaniko_snapshot_mode.set("redo")
        self.kaniko_secrets_mount_dir.set("/run/secrets")

    def finalize(self) -> None:
        if self.cache.get() and not self.push.get() and not self.cache_repo.get():
            self.logger.warning(
                "Disabling cache in Kaniko build %s because it must be combined with push or cache_repo",
                self,
            )
            self.cache.set(False)
        return super().finalize()

    def _render_auth_file(self) -> str:
        """Renders the JSON configuration that will be written to `/kaniko/.docker/config.json`."""
        return json.dumps(
            {
                "auths": {
                    index: {"auth": base64.b64encode(f"{username}:{password}".encode("ascii")).decode("ascii")}
                    for index, (username, password) in self.auth.get().items()
                }
            },
            indent=2,
        )

    def _render_main_script(self, executor_command: list[str]) -> str:
        """Renders the shell script that will be executed in the Kaniko container."""

        docker_config = self._render_auth_file()

        script = []
        script += [
            "mkdir -p /kaniko/.docker",
            "cat << EOF > /kaniko/.docker/config.json",
            docker_config,
            "EOF",
        ]

        if self.secrets:
            script += [f"mkdir -p {shlex.quote(self.kaniko_secrets_mount_dir.get())}"]
            for secret, value in self.secrets.get().items():
                script += [
                    f"echo {shlex.quote(value)} > {shlex.quote(self.kaniko_secrets_mount_dir.get() + '/' + secret)}"
                ]

        script += [" ".join(map(shlex.quote, executor_command))]
        return "\n".join(script)

    def _get_kaniko_executor_command(self, dockerfile: str | None, tar_path: str | None) -> list[str]:
        if tar_path and not self.tags:
            raise ValueError("Need at least one destination (tag) when exporting to an image tarball")
        executor_command = ["/kaniko/executor"]
        executor_command += flatten(("--build-arg", f"{key}={value}") for key, value in self.build_args.get().items())
        cache_repo = self.cache_repo.get()
        if cache_repo:
            executor_command += ["--cache-repo", cache_repo]
        if self.cache.get_or(False):
            executor_command += ["--cache=true"]
        executor_command += flatten(("--destination", destination) for destination in self.tags.get())
        if dockerfile:
            executor_command += ["--dockerfile", dockerfile]
        if not self.push.get():
            executor_command += ["--no-push"]
        executor_command += ["--snapshotMode", self.kaniko_snapshot_mode.get()]
        if self.squash.get():
            executor_command += ["--single-snapshot"]
        if self.kaniko_cache_copy_layers.get():
            executor_command += ["--cache-copy-layers"]
        target = self.target.get()
        if target:
            executor_command += ["--target", target]
        if tar_path:
            executor_command += ["--tarPath", tar_path]
        executor_command += ["--context", self.kaniko_context.get()]
        return executor_command

    def _build(
        self,
        exit_stack: contextlib.ExitStack,
    ) -> None:
        volumes = [f"{self.build_context.get().absolute()}:{self.kaniko_context.get()}"]

        # If the Dockerfile is not relative to the build context, we need to mount it explicitly.
        in_container_dockerfile: str | None = None
        if self.dockerfile.is_filled():
            try:
                in_container_dockerfile = str(
                    self.dockerfile.get().absolute().relative_to(self.build_context.get().absolute())
                )
            except ValueError:
                in_container_dockerfile = "/kaniko/Dockerfile"
                volumes += [f"{self.dockerfile.get().absolute()}:{in_container_dockerfile}"]

        # If the image needs to be loaded into the Docker daemon after building, we need to always
        # export it to a file.
        image_output_file = self.image_output_file.get_or(None)
        if self.load and not image_output_file:
            tempdir = exit_stack.enter_context(tempfile.TemporaryDirectory())
            image_output_file = Path(tempdir) / "image.tgz"

        # Construct the tar path for inside the container.
        tar_path: str | None = None
        if image_output_file:
            volumes += [f"{image_output_file.parent.absolute()}:/kaniko/out"]
            tar_path = f"/kaniko/out/{image_output_file.name}"

        executor_command = self._get_kaniko_executor_command(in_container_dockerfile, tar_path)

        script = self._render_main_script(executor_command)

        result = docker_run(
            image=self.kaniko_image.get(),
            args=["sh", "-c", script],
            entrypoint="",
            remove=True,
            volumes=volumes,
            workdir=self.kaniko_context.get(),
            platform=self.platform.get_or(None),
        )

        if result != 0:
            raise Exception(f"Kaniko build failed with exit code {result}")

        if self.load:
            assert image_output_file is not None, "image_output_file is expected to be set when config.load == True"
            result = docker_load(image_output_file)
            if result != 0:
                raise Exception(f"Docker load failed with exit code {result}")

    def execute(self) -> TaskResult:
        with contextlib.ExitStack() as exit_stack:
            self._build(exit_stack)
        return TaskResult.SUCCEEDED



""" Provides tasks for Rust projects that build using Cargo. """

from __future__ import annotations

import base64
import contextlib
import dataclasses
import enum
import io
import logging
import re
import subprocess as sp
import time
import urllib.parse
from pathlib import Path
from typing import Iterator, List

import tomli
import tomli_w
from kraken.core import Project, Property, Task, TaskResult
from kraken.core.utils import atomic_file_swap, not_none

logger = logging.getLogger(__name__)
CARGO_CONFIG_TOML = Path(".cargo/config.toml")


@dataclasses.dataclass
class PublishTokenType(enum.Enum):
    """Desribes what kind of publish token the registry expects."""

    #: Prefix the password or api_key with `"Bearer "`. Example: Artifactory
    BEARER_TOKEN = enum.auto()

    #: Just take the password or api_key as the publish token as-is. Example: Cloudsmith
    API_KEY = enum.auto()


@dataclasses.dataclass
class Registry:
    """Represents a Cargo registrory."""

    #: Registry name.
    name: str

    #: The index URL.
    index_url: str

    #: The publish token. This may be a templated string that, if applicable, will be constructed from the
    #: :attr:`CargoProjectSettings.auth` credentials if a matching host was found according to *index_url*.
    #: The available variables are `${PASSWORD}` and `${BASIC}`.
    #: Artifactory will expected a `Bearer ${PASSWORD}` while Cloudsmith expects just `${PASSWORD}` (and
    #: `PASSWORD` should be the Cloudsmith API Key).
    publish_token: str

    def __post_init__(self) -> None:
        # Validate the registry name.
        assert re.match(r"^[a-zA-Z0-9_\-]+$", self.name), f"invalid registry name: {self.name!r}"

    def get_publish_token(self, auth: dict[str, tuple[str, str]]) -> str:
        """Replaces variables in the :attr:`publish_token` if necessary and returns it."""

        if not ("${PASSWORD}" in self.publish_token or "${BASIC}" in self.publish_token):
            return self.publish_token

        hostname = not_none(not_none(urllib.parse.urlparse(self.index_url)).hostname)
        if hostname not in auth:
            raise ValueError(
                f"host {hostname!r} has no auth configured in Kraken project settings "
                f"(required for registry {self.name!r})"
            )

        curr_auth = auth[hostname]
        password = curr_auth[1]
        basic = base64.b64encode(f"{curr_auth[0]}:{curr_auth[1]}".encode()).decode("ascii")
        return self.publish_token.replace("${PASSWORD}", password).replace("${BASIC}", basic)


@dataclasses.dataclass
class CargoProjectSettings:
    """Settings for Cargo tasks in a project."""

    #: A dictionary that maps host names to (username, password) tuples.
    auth: dict[str, tuple[str, str]] = dataclasses.field(default_factory=dict)

    #: A dictionary that maps registry index URLs.
    registries: dict[str, Registry] = dataclasses.field(default_factory=dict)

    def add_auth(self, host: str, username: str, password: str) -> CargoProjectSettings:
        self.auth[host] = (username, password)
        return self

    def add_registry(
        self, registry_name: str, index_url: str, publish_token: str = "${PASSWORD}"
    ) -> CargoProjectSettings:
        self.registries[registry_name] = Registry(registry_name, index_url, publish_token)
        return self


def cargo_settings(project: Project | None = None) -> CargoProjectSettings:
    """Creates or gets the cargo project settings for the current project or the given one."""

    if project is None:
        from kraken.api import project as _project

        project = _project

    settings = project.find_metadata(CargoProjectSettings)
    if settings is None:
        settings = CargoProjectSettings()
        project.metadata.append(settings)

    return settings


def _load_gitconfig(file: Path) -> dict[str, dict[str, str]]:
    import configparser

    parser = configparser.RawConfigParser()
    parser.read(file)
    result = dict(parser._sections)  # type: ignore[attr-defined]
    for k in result:
        result[k] = dict(parser._defaults, **result[k])  # type: ignore[attr-defined]
        result[k].pop("__name__", None)
    return result


def _dump_gitconfig(data: dict[str, dict[str, str]]) -> str:
    import configparser

    parser = configparser.RawConfigParser()
    for section, values in data.items():
        parser.add_section(section)
        for key, value in values.items():
            parser.set(section, key, value)
    fp = io.StringIO()
    parser.write(fp)
    return fp.getvalue()


@contextlib.contextmanager
def _cargo_inject_settings(project_dir: Path, certs_dir: Path, settings: CargoProjectSettings) -> Iterator[None]:
    """This context manager injects the HTTP basic authentication headers using an MITM proxy and informs Cargo
    and Git of the HTTP(s) proxy and the self-signed certificates.

    Args:
        project_dir: The directory that contains the Cargo project.
        settings: The project settings that contain the credentials.
    """

    if not settings.auth:
        return

    from kraken.std.cargo.mitm import mitm_auth_proxy

    cargo_config_toml = project_dir / ".cargo" / "config.toml"
    cargo_config = tomli.loads(cargo_config_toml.read_text()) if cargo_config_toml.is_file() else {}

    git_config_file = Path("~/.gitconfig").expanduser()
    git_config = _load_gitconfig(git_config_file) if git_config_file.is_file() else {}

    with contextlib.ExitStack() as exit_stack:

        # Start an HTTP(S) proxy that we can direct Cargo and Git to. We need to give the proxy a little
        # time to come up.
        proxy_url, cert_file = exit_stack.enter_context(mitm_auth_proxy(settings.auth, certs_dir))
        time.sleep(0.5)

        # Temporarily update the Cargo configuration file to inject the HTTP(S) proxy and CA info.
        cargo_registries = cargo_config.setdefault("registries", {})
        for registry in settings.registries.values():
            cargo_registries[registry.name] = {"index": registry.index_url}
        cargo_http = cargo_config.setdefault("http", {})
        cargo_http["proxy"] = proxy_url
        cargo_http["cainfo"] = str(cert_file.absolute())
        logger.info("updating %s", cargo_config_toml)
        fp = exit_stack.enter_context(atomic_file_swap(cargo_config_toml, "w", always_revert=True, create_dirs=True))
        fp.write(tomli_w.dumps(cargo_config))
        fp.close()

        # Temporarily update the Git configuration file to inject the HTTP(S) proxy and CA info.
        git_http = git_config.setdefault("http", {})
        git_http["proxy"] = proxy_url
        git_http["sslCAInfo"] = str(cert_file.absolute())
        logger.info("updating %s", git_config_file)
        fp = exit_stack.enter_context(atomic_file_swap(git_config_file, "w", always_revert=True, create_dirs=True))
        fp.write(_dump_gitconfig(git_config))
        fp.close()

        yield


@contextlib.contextmanager
def _cargo_inject_project_settings(project: Project) -> Iterator[None]:
    settings = cargo_settings(project)
    with _cargo_inject_settings(project.directory, project.context.build_directory / ".mitm-certs", settings):
        yield


class CargoBuildTask(Task):
    """This task runs `cargo build` using the specified parameters. It will respect the authentication
    credentials configured in :attr:`CargoProjectSettings.auth`."""

    args: Property[List[str]]

    def __init__(self, name: str, project: Project) -> None:
        super().__init__(name, project)
        self.args.set([])

    def execute(self) -> TaskResult:
        with _cargo_inject_project_settings(self.project):
            command = ["cargo", "build"] + self.args.get()
            self.logger.info("%s", command)
            result = sp.call(command, cwd=self.project.directory)
            return TaskResult.SUCCEEDED if result == 0 else TaskResult.FAILED


class CargoPublishTask(Task):
    """Publish a Cargo crate."""

    #: The name of the Cargo registry configured in `.cargo/config.yoml` under `[registries]` to publish
    #: the package to.
    registry: Property[str]

    #: Pass the `--allow-dirty` flag to `cargo publish`. This is usually discouraged.
    allow_dirty: Property[bool]

    def __init__(self, name: str, project: Project) -> None:
        super().__init__(name, project)
        self.allow_dirty.set(False)

    def execute(self) -> TaskResult:
        settings = cargo_settings(self.project)
        if self.registry.get() not in settings.registries:
            raise ValueError(f"registry {self.registry.get()!r} is not configured in Kraken project settings")
        registry = settings.registries[self.registry.get()]
        publish_token = registry.get_publish_token(settings.auth)

        with _cargo_inject_project_settings(self.project):
            command = ["cargo", "publish", "--registry", registry.name, "--token", publish_token]
            if self.allow_dirty.get():
                command += ["--allow-dirty"]
            self.logger.info("%s", command)
            result = sp.call(command, cwd=self.project.directory)
            return TaskResult.SUCCEEDED if result == 0 else TaskResult.FAILED



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['std', 'std.cargo', 'std.docker', 'std.generic', 'std.helm']

package_data = \
{'': ['*'], 'std.cargo': ['data/certs/*']}

install_requires = \
['httpx>=0.23.0,<0.24.0',
 'kraken-core>=0.3.0,<0.4.0',
 'proxy.py>=2.4.3,<3.0.0',
 'tomli-w>=1.0.0,<2.0.0',
 'tomli>=2.0.1,<3.0.0']

setup_kwargs = {
    'name': 'kraken-std',
    'version': '0.1.10',
    'description': 'The Kraken standard library.',
    'long_description': '# kraken-std\n\n[![Python application](https://github.com/kraken-build/kraken-std/actions/workflows/python-package.yml/badge.svg)](https://github.com/kraken-build/kraken-std/actions/workflows/python-package.yml)\n[![PyPI version](https://badge.fury.io/py/kraken-std.svg)](https://badge.fury.io/py/kraken-std)\n\nThe Kraken standard library.\n',
    'author': 'Niklas Rosenstein',
    'author_email': 'rosensteinniklas@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



import platform
import shutil
import subprocess as sp
import sys
import tarfile
import tempfile
from pathlib import Path
from typing import List

import httpx
from kraken.core import Project, Property, Task, TaskResult

RELEASE_URL = (
    "https://github.com/estesp/manifest-tool/releases/download/v{VERSION}/binaries-manifest-tool-{VERSION}.tar.gz"
)

# TODO (@NiklasRosenstein): Use existing manifest tool if it exists.
# TODO (@NiklasRosenstein): Ensure manifest-tool has credentials to push to the target


class ManifestToolPushTask(Task):
    """A task that uses `manifest-tool` to combine multiple container images from different platforms into a single
    multi-platform manifest.

    For more information on `manifest-tool`, check out the GitHub repository:

    https://github.com/estesp/manifest-tool/
    """

    #: The Docker platforms to create the manifest for.
    platforms: Property[List[str]]

    #: A Docker image tag that should contain the variables `OS`, `ARCH` and `VARIANT`.
    template: Property[str]

    #: The image ID to push the Docker image to.
    target: Property[str]

    #: Prefer the local version of the tool if available. Default is `true`.
    manifest_tool_local: Property[bool]

    #: The tool version to use. The appropriate release will be downloaded from Github.
    manifest_tool_version: Property[str]

    def __init__(self, name: str, project: Project) -> None:
        super().__init__(name, project)
        self.manifest_tool_local.set(True)
        self.manifest_tool_version.set("2.0.4")

    def fetch_manifest_tool(self) -> Path:
        """Fetches the manifest tool binary that is appropriate for the current platform."""

        if self.manifest_tool_local.get():
            path = shutil.which("manifest-tool")
            if path is not None:
                self.logger.info("using %s", path)
                return Path(path)

        version = self.manifest_tool_version.get()
        manifest_tool_dir = self.project.context.build_directory / ".downloads" / f"manifest-tool-{version}"
        if not manifest_tool_dir.is_dir():
            download_url = RELEASE_URL.format(VERSION=version)
            self.logger.info("downloading manifest-tool release v%s (%s) ...", version, download_url)
            with httpx.stream(
                "GET", download_url, follow_redirects=True
            ) as fp, tempfile.TemporaryDirectory() as tempdir:
                fp.raise_for_status()
                archive = Path(tempdir) / download_url.rpartition("/")[-1]
                with archive.open("wb") as dst:
                    for chunk in fp.iter_bytes():
                        dst.write(chunk)
                with tarfile.open(archive, mode="r:gz") as tf:
                    tf.extractall(manifest_tool_dir)

        filename = f"manifest-tool-{sys.platform}-{platform.machine()}"
        binary = manifest_tool_dir / filename
        if not binary.is_file():
            raise RuntimeError("unable to construct valid path to binary for downloaded manifest-tool release")

        self.logger.info("using %s", binary)
        return binary

    def execute(self) -> TaskResult:
        binary = self.fetch_manifest_tool()
        command = [
            str(binary),
            "push",
            "from-args",
            "--platforms",
            ",".join(self.platforms.get()),
            "--template",
            self.template.get(),
            "--target",
            self.target.get(),
        ]
        self.logger.info("%s", command)
        result = sp.call(command)
        if result != 0:
            return TaskResult.FAILED
        return TaskResult.SUCCEEDED
